# Copyright © 2024 Luxembourg Institute of Science and Technology
# Licensed under the MIT License (see LICENSE file for details).
# Authors: [Sabina Bednářová, Thomas Gibon]

import shutil
import pandas as pd
from datetime import datetime
from pathlib import Path
import requests
import json
import os
import datetime
import time
import appdirs
import sys
import tempfile, subprocess
from shrecc.treatment import save_to_pickle, load_from_pickle


def get_prod(start, end, country, cumul=False, rolling=False):
    """
    Downloads production data from the Energy Charts API. Gets called from `get_data()`.

    Args:
        start (int): Start of the download period (output of `year_to_unix()`) in unix seconds.
        end (int): End of the download period (output of `year_to_unix()`) in unix seconds.
        country (list of str): The country for which data needs to be downloaded.
        cumul (bool): If True, calculate the cumulative sum of production.
        rolling (bool): If True, calculate the rolling average of production.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame, np.ndarray]: A tuple containing:
            - A dataframe of production.
            - A dataframe of load.
            - An array of all available technologies.
    """
    s = requests.Session()
    url = f"https://api.energy-charts.info/public_power?country={country}&start={start}&end={end}"
    r = s.get(url)
    s.close()
    if r.status_code == 404:
        print("File not found!")
        return None, None, None
    if r.status_code == 400:
        print("Bad request! Country not found.")
        return None, None, None
    response = json.loads(r.text)
    techs = []
    prod = []
    for r in response["production_types"]:
        try:
            techs.append(r["name"])
            prod.append(r["data"])
        except TypeError:
            print("Somethings wrong")
    ticks = [
        pd.to_datetime(d, unit="s", origin="unix") for d in response["unix_seconds"]
    ]
    load_dict = {"en": "Load"}
    print(url)
    prod_df = pd.DataFrame(data=prod, index=techs, columns=ticks).T
    col_exclude = ["Residual load", "Renewable Share"]
    for col in col_exclude:
        if col in prod_df.columns:
            prod_df.drop(col, axis=1, inplace=True)
    if rolling:
        prod_df = prod_df.rolling(rolling).sum()
    if cumul:
        prod_df = prod_df.cumsum()
    try:
        load_df = prod_df[load_dict["en"]]
        prod_df.drop(load_dict["en"], axis=1, inplace=True)
    except Exception as e:
        print(e)
        load_df = None
    print("...production for " + country + " OK.")
    return prod_df, load_df, techs


def get_trade(start, end, country):
    """
    Downloads trade data from the Energy Charts API. Gets called from `get_data()`.

    Args:
        start (int): Start of the download period (output of `year_to_unix()`) in unix seconds.
        end (int): End of the download period (output of `year_to_unix()`) in unix seconds.
        country (list of str): The country for which data needs to be downloaded.

    Returns:
        Tuple[pd.DataFrame, np.ndarray]: A tuple containing:
            - A dataframe of trades between countries.
            - An array of all available regions.
    """
    s = requests.Session()
    url = (
        f"https://api.energy-charts.info/cbpf?country={country}&start={start}&end={end}"
    )
    r = s.get(url)
    s.close()
    if r.status_code == 404:
        print("File not found!")
        return None, None
    response = json.loads(r.text)
    ticks = [
        pd.to_datetime(d, unit="s", origin="unix") for d in response["unix_seconds"]
    ]
    trade = []
    regions = []
    for r in response["countries"]:
        trade.append(r["data"])
        regions.append(r["name"])
    print(url)
    trade_df = pd.DataFrame(data=trade, index=regions, columns=ticks).T
    print("...trade for " + country + " OK.")
    return trade_df, regions


def get_data(year, root=None, max_retries=3, retry_delay=5):
    """
    Main function for downloading data.

    Args:
        year (int): The selected year for which data is to be downloaded, e.g., 2023.
        root (Path): location of the data.
        max_retries (int): The maximum number of retries for each country download in case of problems.
        retry_delay (int): The delay in seconds between retries.

    Returns:
        pd.DataFrame: A dataframe containing both production and trade data for all countries in the selected year.
    """
    if root is None:
        package_name = sys.modules["__main__"].__package__
        if package_name is None:
            package_name = "shrecc"
        root = Path(appdirs.user_data_dir(package_name))
    elif isinstance(root, str):
        root = Path(root)
    countries = [
        "AL",
        "AM",
        "AT",
        "AZ",
        "BA",
        "BE",
        "BG",
        "BY",
        "CH",
        "CY",
        "CZ",
        "DE",
        "DK",
        "EE",
        "ES",
        "FI",
        "FR",
        "GE",
        "GR",
        "HR",
        "HU",
        "IE",
        "IT",
        "LT",
        "LU",
        "LV",
        "MD",
        "ME",
        "MK",
        "MT",
        "NIE",
        "NL",
        "NO",
        "PL",
        "PT",
        "RO",
        "RS",
        "RU",
        "SE",
        "SK",
        "SI",
        "TR",
        "UA",
        "UK",
        "XK",
    ]
    filename = Path(root) / "data" / f"{year}" / f"prod_and_trade_data_{year}.pkl"
    filename.parent.mkdir(parents=True, exist_ok=True)
    start, end = year_to_unix(year)
    if filename.exists():
        data = load_from_pickle(filename)
        print("API data loaded successfully.")
    else:
        data = dict()
        for country in countries:
            country = country.lower()
            print(country)
            for attempt in range(max_retries):
                try:
                    prod_df, load_df, _ = get_prod(
                        start=start, end=end, country=country, cumul=False, rolling=1
                    )

                    trade_df, _ = get_trade(
                        start=start,
                        end=end,
                        country=country,
                    )

                    data[country] = {
                        "production mix": prod_df,
                        "load": load_df,
                        "trade": trade_df,
                    }
                    break
                except requests.ConnectionError as e:
                    print(
                        f"Network error: {e}, retrying {attempt + 1}/{max_retries}..."
                    )
                    time.sleep(retry_delay)
                except Exception as e:
                    print(f"Error: {e}, retrying {attempt + 1}/{max_retries}...")
                    time.sleep(retry_delay)
            else:
                print(f"Failed to fetch data for {country}.")
        save_to_pickle(data, filename)
    data_df = cleaning_data(data, root)
    return data_df


def year_to_unix(year):
    """
    Converts a year to Unix timestamps representing the start and end of the year.

    Args:
        year (int): The selected year, passed from `get_data()`.

    Returns:
        Tuple[int, int]: A tuple containing:
            - The start of the year in Unix seconds.
            - The end of the year in Unix seconds.
    """
    start_of_year = datetime.datetime(year, 1, 1, 0, 0)
    end_of_year = datetime.datetime(year, 12, 31, 23, 59)
    start_unix = int(time.mktime(start_of_year.timetuple()))
    end_unix = int(time.mktime(end_of_year.timetuple()))
    return start_unix, end_unix


def cleaning_data(data, root):
    """
    Cleans the data and adds missing countries. Note that missing countries need to be manually added to `country_codes`.
    Gets called from `get_data()`.

    Args:
        data (pd.DataFrame): The dataframe containing production and trade data.
        root (Path): location of the data.

    Returns:
        pd.DataFrame: A dataframe with missing countries added.
    """
    techs = []
    partners = []
    for country, datasets in data.items():
        try:
            techs.extend(datasets["production mix"].columns)
            partners.extend(datasets["trade"].columns)
        except:
            pass
    filename = Path(root) / "data" / f"generation_units_by_country.csv"
    if filename.exists():
        gen_units_per_country = pd.read_csv(filename, index_col=1)["short"]
    country_codes = {
        p: gen_units_per_country.loc[p]
        for p in set(partners)
        if p in gen_units_per_country.index
    }
    missing_countries = set(partners) ^ country_codes.keys()
    country_codes = {
        **country_codes,
        **{
            "Armenia": "AM",
            "Azerbaijan": "AZ",
            "Cyprus": "CY",  # Does not appear?
            "Ireland": "IE",
            "Malta": "MT",
            "North Macedonia": "MK",
            "Serbia": "RS",
            "Slovakia": "SK",
        },
    }
    data_clean = dict()
    filename = Path(root) / "data" / f"techs_agg.json"
    if filename.exists():
        with open(filename, "r") as f:
            techs_agg = json.load(f)
    agg_dict = {
        "production mix": techs_agg,
        "trade": country_codes,
        "load": {"Load": "load"},
    }
    scale_dict = {"production mix": 1, "trade": 1000, "load": 1}

    for country in data.keys():
        data_clean[country.upper()] = dict()
        for k, v in data[country].items():
            if type(v) is pd.DataFrame:  # axis = 1 will soon be depreciated
                grouped = v.T.groupby(agg_dict[k]).sum().T
                grouped.index = pd.to_datetime(grouped.index)
                data_clean[country.upper()][k] = (
                    grouped.resample("h").mean() * scale_dict[k]
                )

            elif type(v) is pd.Series:
                v.index = pd.to_datetime(v.index)
                data_clean[country.upper()][k] = v.resample("h").mean() * scale_dict[k]

    data_clean = {k: v for k, v in data_clean.items() if v != {}}
    P = pd.concat(
        [
            pd.concat(
                {country: pd.concat(data_clean[country], axis=1)},
                axis=1,
                names=["country", "type", "source"],
            )
            for country in data_clean.keys()
        ],
        axis=1,
    )
    return P


def download_shrecc_data(
    repo_url="https://git.list.lu/shrecc_project/shrecc_data/",
    destination_directory=None,
    branch="main",
) -> None:
    """
    Utility function to download the whole pre-calculate dataset.

    Args:
        repo_url (str): the default location of the repository with the data
        destination_directory: destination of data download
        branch (str): branch of the repository

    """
    clone_last_commit(repo_url, destination_directory, branch)


def clone_last_commit(repo_url, destination_directory=None, branch="main"):
    """
    Clone the latest commit from a specified branch of a Git repository.

    This function performs a shallow clone of the latest commit from a specified
    branch of a Git repository and copies the repository contents to the destination
    directory, excluding the `.git` directory, `.gitattributes`, and `README.md`.

    Args:
        repo_url (str): The URL of the Git repository.
        destination_directory (str): The directory where the repository contents will be copied.
                                     If not specified, it will be stored in the SHRECC module appdir.
        branch (str): The branch to clone. Defaults to 'main'.

    Raises:
        subprocess.CalledProcessError: If the git clone command fails.
    """
    if destination_directory is None:
        package_name = sys.modules["__main__"].__package__
        if package_name is None:
            package_name = "shrecc"
        destination_directory = appdirs.user_data_dir(package_name)

    os.makedirs(destination_directory, exist_ok=True)
    ignore_from_download = [".git", ".gitattributes", "README.md"]

    with tempfile.TemporaryDirectory() as tmpdirname:
        try:
            subprocess.run(
                [
                    "git",
                    "clone",
                    "--depth",
                    "1",
                    "--branch",
                    branch,
                    repo_url,
                    tmpdirname,
                ],
                check=True,
            )
            print(
                f"Repository cloned successfully into temporary directory '{tmpdirname}'."
            )
            print(f"Moving downloaded contents to '{destination_directory}'.")
            for item in os.listdir(tmpdirname):
                s = os.path.join(tmpdirname, item)
                d = os.path.join(destination_directory, item)
                if os.path.exists(s) and item in ignore_from_download:
                    continue
                if os.path.isdir(s):
                    if not os.path.exists(d):
                        shutil.copytree(s, d)
                    else:
                        print(f"Directory '{d}' already exists, skipping.")
                else:
                    if not os.path.exists(d):
                        shutil.copy2(s, d)
                    else:
                        print(f"File '{d}' already exists, skipping.")
            print(
                f"Repository contents copied successfully to '{destination_directory}'."
            )

        except subprocess.CalledProcessError as e:
            print(f"Failed to clone repository. Error: {e}")
            print("Please make sure you have git installed.")



# Copyright © 2024 Luxembourg Institute of Science and Technology
# Licensed under the MIT License (see LICENSE file for details).
# Authors: [Sabina Bednářová, Thomas Gibon]


"""shrecc"""

__all__ = (
    "__version__",
    "create_database",
    "filt_cutoff",
    "get_data",
    "data_processing",
    "download_shrecc_data",
)
from .database import create_database, filt_cutoff
from .download import get_data, download_shrecc_data
from .treatment import data_processing

__version__ = "0.0.1"



# Copyright © 2024 Luxembourg Institute of Science and Technology
# Licensed under the MIT License (see LICENSE file for details).
# Authors: [Sabina Bednářová, Thomas Gibon]

from pathlib import Path
import sys
import appdirs
import numpy as np
import pandas as pd
import bw2data as bd
from bw2data.query import Query, Filter
from shrecc.treatment import save_to_pickle, load_from_pickle
from datetime import datetime


def filt_cutoff(
    countries,
    times=0,
    general_range=0,
    refined_range=0,
    freq=0,
    cutoff=1e-3,
    include_cutoff=True,
    root=None,
):
    """
    Filters data based on selected countries and times (either one-off, a range, or periodical range). Gets called from `create_database()`.

    Args:
        year (int): Selected year of the downloaded data.
        countries (list of str): Countries selected by the user for their database.
            E.g. countries=['FR', 'DE'].
        times (list of str): Selecting one specific time, e.g. times = ['2023-06-16 8:00:00', '2023-06-16 22:00:00'].
            Can be applied alone.
        general_range (list of str): Selecting a general range, e.g. for the month of June
            general_range = ['2023-06-01 01:00:00', '2023-06-30 23:00:00']. Can be applied alone.
        refined_range (list of int): Refining range of general range, e.g. mornings of June (previously selected in general_range):
            refined_range = [8, 9, 10, 11]. Can only be applied with general_range.
        freq (str): Days to be included, e.g. freq='D' selects calendar days,
            see https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases.
        cutoff (float): Cutoff value for technology values.
        include_cutoff (bool): If True, cutoff is applied and summed at the end to create a new technology "The rest".
            If False, cutoff is applied but new technology not created.
        root (Path): location of the data.

    Returns:
        pd.DataFrame: The filtered dataframe.
    """
    if root is None:
        package_name = sys.modules["__main__"].__package__
        if package_name is None:
            package_name = "shrecc"
        root = Path(appdirs.user_data_dir(package_name))
    elif isinstance(root, str):
        root = Path(root)
    year = datetime.strptime(general_range[0], "%Y-%m-%d %H:%M:%S").year
    dataframe = tech_mapping(year, root=root)
    print("Filtering dataframe...")
    dataframe = dataframe.droplevel("source", axis=1)
    dataframe = filter_by_countries(dataframe, countries)
    if times:
        dataframe = filter_by_times(dataframe, times)
    if general_range:
        dataframe = filter_by_range(dataframe, general_range, refined_range, freq)
    dataframe = apply_cutoff(dataframe, cutoff, include_cutoff)
    print("Dataframe filtered.")
    return dataframe


def load_mapping_data(path_to_data):
    """
    Load the mapping data from an Excel file.

    Args:
        path_to_data (str or Path): The path to the directory containing the mapping data.

    Returns:
        pd.DataFrame: A DataFrame containing the technology mapping data from the Excel file.
    """
    return pd.read_csv(
        path_to_data / "el_map_all_norm.csv",
        index_col=[0, 1, 2, 3],
        header=[0, 1],
    )


def load_time_series_data(path_to_data, year):
    """
    Load the time series data from a pickle file and format it as a DataFrame.

    Args:
        path_to_data (str or Path): The path to the directory containing the time series data.
        year (int): The year corresponding to the time series data.

    Returns:
        pd.DataFrame: A DataFrame containing the time series data, with levels reordered and sorted.
    """
    filename = path_to_data / f"{year}" / f"indices_{year}.pkl"
    indices = load_from_pickle(filename)
    Z_cons_sp = load_from_pickle(path_to_data / f"{year}" / f"Z_cons_{year}.pkl")
    Z_cons = pd.DataFrame(
        np.float32(Z_cons_sp.todense()),
        index=indices["index"],
        columns=indices["columns"],
    )
    return Z_cons.reorder_levels(["time", "source", "country"], axis=1).sort_index(
        axis=1
    )


def prepare_consumption_data(Z_cons):
    """
    Prepare the consumption data by removing trade data and adjusting indices.

    Args:
        Z_cons (pd.DataFrame): The original consumption data DataFrame.

    Returns:
        pd.DataFrame: The prepared consumption data, with the trade data removed and indices swapped.
    """
    Z_cons = Z_cons.sort_index()
    Z_cons_to_multiply = Z_cons.drop("trade", axis=0).copy()
    Z_cons_to_multiply.index.names = ["source", "geography_mix"]
    return Z_cons_to_multiply.swaplevel()


def apply_mapping(Z_cons_to_multiply, el_map_all_norm):
    """
    Apply the technology mapping to the consumption data.

    Args:
        Z_cons_to_multiply (pd.DataFrame): The consumption data to be mapped.
        el_map_all_norm (pd.DataFrame): The normalized mapping data.

    Returns:
        pd.DataFrame: The resulting DataFrame after applying the technology mapping.
    """
    
    # Changing data types to float32 significantly reduces the product time
    el_map_to_multiply = el_map_all_norm.reindex(Z_cons_to_multiply.index, axis=1).astype('float32')
    el_map_to_multiply["NIE"] = el_map_all_norm["GB"][el_map_to_multiply["NIE"].columns].astype('float32')
    el_map_to_multiply["UK"] = el_map_all_norm["GB"][el_map_to_multiply["UK"].columns].astype('float32')

    LCI_cons = el_map_to_multiply.dot(Z_cons_to_multiply)

    if LCI_cons.isna().sum().sum():
       return LCI_cons.fillna(0)
        
    return LCI_cons


def tech_mapping(year, root=None):
    """
    Main function to map the technologies and scale them to 1 kWh.

    Args:
        year (int): The year corresponding to the data.
        root (path): Root directory.

    Returns:
        pd.DataFrame: A DataFrame with the scaled technology mappings.
    """
    print("Mapping technologies...")
    el_map_all_norm = load_mapping_data(Path(root / "data"))
    Z_cons = load_time_series_data(Path(root / "data"), year)
    filename = Path(root / "data" / f"{year}" / f"indices_{year}.pkl")
    Z_cons_to_multiply = prepare_consumption_data(Z_cons)
    filename = Path(root) / "data" / f"{year}" / f"LCI_cons_scaled_{year}.pkl"
    if filename.exists():
        LCI_cons_scaled = load_from_pickle(filename)
    else:
        LCI_cons = apply_mapping(Z_cons_to_multiply, el_map_all_norm)
        sum = LCI_cons.sum()
        filename = Path(root) / "data" / f"{year}" / f"Z_load_{year}.pkl"
        load = load_from_pickle(filename)
        load_difference_row = (
            "RER",
            "electricity, high voltage, European attribute mix",
            "electricity, high voltage",
            "kWh",
        )

        # LCI_cons.loc[load_difference_row] = 0

        load_stacked = load.stack()
        load_stacked.index.names = ["country", "time"]
        load_stacked = load_stacked[load_stacked != 0]
        merged = load_stacked.reset_index(name="load_value").merge(
            sum.reset_index(), how="inner", on=["time", "country"]
        )
        merged.rename(columns={0: "sum"}, inplace=True)
        merged["difference"] = (merged["load_value"] - merged["sum"]) 
        merged = merged[merged["difference"] > 0]
        merged.set_index(["time","source","country"], inplace=True)

        LCI_cons.sort_index(inplace=True)  # Sort rows
        LCI_cons.sort_index(axis=1, inplace=True)

        # LCI_cons.loc[load_difference_row, merged.index] = merged["difference"]

        LCI_cons.loc[load_difference_row] = merged["difference"]
        LCI_cons.loc[load_difference_row].fillna(0, inplace=True)

        # for _, row in merged.iterrows():
        #     time = row["time"]
        #     country = row["country"]
        #     difference = row["difference"]
        #     lci_col = (time, "trade", country)
        #     if lci_col in LCI_cons.columns:
        #         LCI_cons.loc[load_difference_row, lci_col] += difference
        
        LCI_cons_scaled = LCI_cons / LCI_cons.sum()
        save_to_pickle(
            LCI_cons_scaled, Path(root) / "data" / f"LCI_cons_scaled_{year}.pkl"
        )
    print("Technologies mapped.")
    return LCI_cons_scaled


def filter_by_countries(dataframe, countries):
    """
    Filter the dataframe by selected countries.

    Args:
        dataframe (pd.DataFrame): The original dataframe containing data for multiple countries.
        countries (list of str): A list of country codes to filter by.

    Returns:
        pd.DataFrame: A dataframe filtered by the specified countries.
    """
    if "country" in dataframe.columns.names:
        return dataframe.loc[
            :, dataframe.columns.get_level_values("country").isin(countries)
        ]
    else:
        print("Couldnt find country to filter")


def filter_by_times(dataframe, times):
    """
    Filter the dataframe by specific times.

    Args:
        dataframe (pd.DataFrame): The original dataframe containing data for multiple times.
        times (list of str): A list of specific times to filter by.

    Returns:
        pd.DataFrame: A dataframe filtered by the specified times.
    """
    return dataframe.loc[:, dataframe.columns.get_level_values("time").isin(times)]


def filter_by_range(dataframe, general_range, refined_range, freq):
    """
    Filter the dataframe by a general time range and optionally by a refined time range.

    Args:
        dataframe (pd.DataFrame): The original dataframe containing data.
        general_range (list of str): The start and end of the general range to filter by (e.g., ['2023-06-01', '2023-06-30']).
        refined_range (list of int): A list specifying the refined range of hours to filter within the general range.
        freq (str): The frequency for generating timestamps (e.g., 'D' for daily).

    Returns:
        pd.DataFrame: A dataframe filtered by the specified time range and refined range.
    """
    df_filt = dataframe.loc[:, general_range[0] : general_range[1]]
    if refined_range and len(refined_range) > 1:
        timestamp = pd.date_range(
            start=general_range[0], end=general_range[1], freq=freq, normalize=True
        ) + pd.Timedelta(hours=refined_range[0])
        timestamps_range = timestamp[
            (timestamp.hour >= refined_range[0]) & (timestamp.hour <= refined_range[1])
        ]
        df_filt = df_filt.loc[
            :, df_filt.columns.get_level_values("time").isin(timestamps_range)
        ]
    elif refined_range and len(refined_range) == 1:
        timestamp = pd.date_range(
            start=general_range[0], end=general_range[1], freq=freq, normalize=True
        ) + pd.Timedelta(hours=refined_range[0])
        df_filt = df_filt.loc[
            :, df_filt.columns.get_level_values("time").isin(timestamp)
        ]
    return df_filt.T.groupby(level="country").mean().T


def apply_cutoff(df_filt, cutoff, include_cutoff):
    """
    Apply a cutoff value to filter out smaller values in the dataframe and optionally include a "rest" category.

    Args:
        df_filt (pd.DataFrame): The filtered dataframe.
        cutoff (float): The cutoff value for technology values.
        include_cutoff (bool): If True, sums values below cutoff and includes them as a new technology "The rest".

    Returns:
        pd.DataFrame: A dataframe with values below the cutoff set to zero, optionally including a "rest" category.
    """
    sums = []
    for col in df_filt.columns:
        select = df_filt.loc[:, col]
        sum_smaller_than_cutoff = select[select.le(cutoff)].sum()
        sums.append(sum_smaller_than_cutoff)
    df_filt[df_filt < cutoff] = 0
    if include_cutoff:
        df_filt.loc[
            (
                "RER",
                "market group for electricity, high voltage",
                "electricity, high voltage",
                "kWh",
            ),
            :,
        ] = sums
    return df_filt


def setup_database(project_name, db_name):
    """
    Sets up the BW2 database for the given project.

    Args:
        project_name (str): The name of the BW project.
        db_name (str): The name of the BW database to set up.

    Returns:
        bd.Database: The newly registered BW2 database.
    """
    bd.projects.set_current(project_name)
    if db_name in bd.databases:
        del bd.databases[db_name]
        bd.projects.purge_deleted_directories()
    elec_db = bd.Database(db_name)
    elec_db.register()
    return elec_db


def map_known_inputs(eidb_name, dataframe_filt):
    """
    Maps known inputs from the ecoinvent database to the filtered dataframe.

    Args:
        eidb_name (str): The name of the ecoinvent database in the BW project.
        dataframe_filt (pd.DataFrame): The filtered dataframe containing technology data.

    Returns:
        dict: A dictionary mapping known inputs to their corresponding entries in the ecoinvent database.
    """
    ei_db = bd.Database(eidb_name)
    ei_db_data = ei_db.load()
    known_inputs = {}
    for idx in dataframe_filt.index:
        loc, name, prod, unit = idx
        # Region names in ENTSOE and ecoinvent don't exactly match
        # Only UK seems concerned but consider using a dictionary
        if loc == "UK":
            loc = "GB"
        q = Query()
        filter_name = Filter("name", "is", name)
        filter_loc = Filter("location", "is", loc)
        filter_unit = Filter("unit", "is", "kilowatt hour")
        q.add(filter_name)
        q.add(filter_loc)
        q.add(filter_unit)
        results = q(ei_db_data)
        if len(results) == 1:
            known_inputs[(loc, name, unit)] = list(results).pop()
        else:
            print("Couldnt find activity:" + name + ", " + loc)
    network = get_network_activities()
    known_inputs_network = {}
    for act in network:
        loc = act["loc"]
        name = act["name"]
        q = Query()
        filter_name = Filter("name", "is", name)
        filter_loc = Filter("location", "is", loc)
        q.add(filter_name)
        q.add(filter_loc)
        results = q(ei_db_data)
        if len(results) == 1:
            known_inputs_network[(loc, name)] = list(results).pop()
        else:
            print("Couldnt find activity:" + name)
    return known_inputs, known_inputs_network


def get_network_activities():
    activities = [
        "market for distribution network, electricity, low voltage",
        "market for transmission network, electricity, medium voltage",
        "market for sulfur hexafluoride, liquid",
        "market for transmission network, electricity, high voltage direct current aerial line",
        "market for transmission network, electricity, high voltage direct current land cable",
        "market for transmission network, electricity, high voltage direct current subsea cable",
        "transmission network construction, electricity, high voltage",
    ]
    locations = [
        "GLO",
        "GLO",
        "RER",
        "GLO",
        "GLO",
        "GLO",
        "CH",
    ]
    values = [
        8.679076855e-8,
        1.86646177072e-8,
        1.27657893204915e-7,
        8.38e-9,
        3.47e-10,
        5.66e-10,
        6.58e-9,
    ]
    network_act = [
        {"name": activity, "loc": location, "val": value}
        for activity, location, value in zip(activities, locations, values)
    ]
    return network_act


def create_activity_dict(dataframe_filt, known_inputs, known_inputs_network, db_name):
    """
    Creates a dictionary of activities for the BW database based on the filtered dataframe and known inputs.

    Args:
        dataframe_filt (pd.DataFrame): The filtered dataframe containing technology data.
        known_inputs (dict): A dictionary mapping known inputs to ecoinvent database entries.
        db_name (str): The name of the BW database.

    Returns:
        dict: A dictionary containing activities to be written to the BW2 database.
    """
    activities = {}
    for i, col in enumerate(dataframe_filt.columns):
        if dataframe_filt.columns.nlevels > 1:
            time, country = col
            activity = f"{time} Electricity mix"
            name = f"{activity} in {country}"
        else:
            country = col
            name = f"Electricity mix in {country}"
        code = f"electricity {i}"
        act = {
            "name": name,
            "unit": "kWh",
            "code": code,
            "location": str(country),
            "reference product": "Electricity mix",
            "type": "process",
            "exchanges": [],
        }
        for idx in dataframe_filt.index:
            source, exch_name, prod, unit = idx
            if float(dataframe_filt.loc[idx, col]) != 0:
                exchange = known_inputs.get((source, exch_name, unit))
                if exchange:
                    new_exchange = {
                        "input": exchange,
                        "amount": float(dataframe_filt.loc[idx, col]),
                        "type": "technosphere",
                    }
                    act["exchanges"].append(new_exchange)
        network = get_network_activities()
        specific_network = [
            "market for transmission network, electricity, high voltage direct current land cable",
            "market for transmission network, electricity, high voltage direct current subsea cable",
            "transmission network construction, electricity, high voltage",
        ]
        land_cable = [
            "FR",
            "IT",
            "GR",
            "DK",
            "AT",
            "BE",
            "EE",
            "ES",
            "FI",
            "HR",
            "IT",
            "LT",
            "LV",
            "ME",
            "MT",
            "NO",
            "SE",
            "TR",
            "UK",
        ]
        subsea_cable = [
            "FR",
            "IT",
            "GR",
            "DK",
            "EE",
            "ES",
            "FI",
            "HR",
            "IT",
            "LT",
            "ME",
            "MT",
            "NO",
            "PL",
            "UK",
        ]
        for exch_net in network:
            if exch_net["name"] not in specific_network:
                exchange = known_inputs_network.get((exch_net["loc"], exch_net["name"]))
                if exchange:
                    new_exchange = {
                        "input": exchange,
                        "amount": float(exch_net["val"]),
                        "type": "technosphere",
                    }
                    act["exchanges"].append(new_exchange)
            if (
                country in land_cable and exch_net["name"] == specific_network[0]
            ):  # writing the land cable
                exchange = known_inputs_network.get((exch_net["loc"], exch_net["name"]))
                if exchange:
                    new_exchange = {
                        "input": exchange,
                        "amount": float(exch_net["val"]),
                        "type": "technosphere",
                    }
                    act["exchanges"].append(new_exchange)
            if (
                country in subsea_cable and exch_net["name"] == specific_network[1]
            ):  # writing the subsea cable
                exchange = known_inputs_network.get((exch_net["loc"], exch_net["name"]))
                if exchange:
                    new_exchange = {
                        "input": exchange,
                        "amount": float(exch_net["val"]),
                        "type": "technosphere",
                    }
                    act["exchanges"].append(new_exchange)
            if country == "CH" and exch_net["name"] == specific_network[2]:
                exchange = known_inputs_network.get((exch_net["loc"], exch_net["name"]))
                if exchange:
                    new_exchange = {
                        "input": exchange,
                        "amount": float(exch_net["val"]),
                        "type": "technosphere",
                    }
                    act["exchanges"].append(new_exchange)
        activities[(db_name, code)] = act
    return activities


def create_database(dataframe_filt, project_name, db_name, eidb_name, network="True"):
    """
    Creates an "ecoinvent-like" BW database based on a previously filtered dataframe.

    Args:
        dataframe_filt (pd.DataFrame): Scaled and filtered dataframe.
        project_name (str): BW project name to which the database will be saved.
        db_name (str): Name of the BW database to be created.
        eidb_name (str): Name of the ecoinvent database. Must be the same as in the BW project.

    Returns:
        None
    """
    elec_db = setup_database(project_name, db_name)
    if network == "True":
        known_inputs, known_inputs_network = map_known_inputs(eidb_name, dataframe_filt)
        activities = create_activity_dict(
            dataframe_filt, known_inputs, known_inputs_network, db_name
        )
    else:
        known_inputs, _ = map_known_inputs(eidb_name, dataframe_filt)
        activities = create_activity_dict(dataframe_filt, known_inputs, _, db_name)
    elec_db.write(activities)



# Copyright © 2024 Luxembourg Institute of Science and Technology
# Licensed under the MIT License (see LICENSE file for details).
# Authors: [Sabina Bednářová, Thomas Gibon]

import pickle
import pandas as pd
import numpy as np
from pathlib import Path
import appdirs
import sys
import scipy.sparse as sp
from pypardiso import spsolve
from scipy.sparse import coo_array, csr_matrix, identity, csc_matrix
from scipy.linalg import block_diag


def data_processing(data_df, year, root=None):
    """
    Processes data, adds missing countries, and correctly divides them between consumption and demand.

    Args:
        data_df (pd.DataFrame): All of the downloaded data for the selected year (from `get_data`).
        year (int): The selected year, e.g., 2023.
        root (Path): location of the data.

    Returns:
        None: Doesn't return anything, but saves files to the directory 'data'.
    """
    if root is None:
        package_name = sys.modules["__main__"].__package__
        if package_name is None:
            package_name = "shrecc"
        root = Path(appdirs.user_data_dir(package_name))
    elif isinstance(root, str):
        root = Path(root)
    print("Processing data...")

    # If pandas is recent enough, use the future_stack argument in stack
    params = {'level':0}
    if pd.__version__ >= "2.1.0":
        params.update({'future_stack':True})

    Z = (
        data_df.T.unstack("country")
        .drop(
            [
                col
                for col in [("load", "Load"), ("Load", "Load")]
                if col in data_df.T.index
            ],
            errors="ignore",
        )
        .fillna(0)
        .stack(**params)
    )
    Z_trade = pd.concat(
        [Z.loc["trade"]], keys=["trade"], names=["source", "country", "time"]
    )

    missing_countries = set(Z.loc["trade"].columns.values) ^ set(
        Z.loc["trade"].index.get_level_values(0).unique()
    )
    missing_countries.add("CY")
    missing_index = missing_countries - set(Z_trade.index.get_level_values(1).unique())
    missing_columns = missing_countries - set(Z_trade.columns)

    Z_trade[list(missing_columns)] = 0
    t_index = Z.index.levels[2]
    n_t = len(t_index)
    n_c = len(Z_trade.columns)

    to_append_index = pd.DataFrame(
        np.zeros([len(missing_index) * n_t, n_c]),
        index=pd.MultiIndex.from_product([["trade"], missing_index, t_index]),
        columns=Z_trade.columns,
    )

    Z_trade = (
        pd.concat([Z_trade, to_append_index], axis=0).sort_index().sort_index(axis=1)
    )
    Z_trade_sq = Z_trade.unstack()
    Z_trade_sq.index.names = ["source", "country"]
    Z_trade_sq.columns.names = ["country", "time"]
    cols_trade = Z_trade_sq.columns

    Z[list(missing_columns)] = 0
    c_index = Z_trade.columns
    p_index = (
        Z.loc["production mix"]
        .drop(["Import balance (market)"])
        .index.get_level_values(0)
        .unique()
    )  # Sometimes,
    # 'Day Ahead Auction' needs to be dropped as well.
    n_p = len(p_index)
    Zu = Z.unstack()
    Z_prod_diag = pd.DataFrame(
        block_diag(
            *[
                Zu.loc["production mix", c].drop(["Import balance (market)"])
                for c in c_index
            ]
        ),
        index=pd.MultiIndex.from_product([c_index, p_index]),
        columns=cols_trade,
    ).swaplevel(0, 1)
    Z_trade_sq[Z_trade_sq < 0] = 0
    Zx = pd.concat(
        [pd.concat([Z_prod_diag, Z_trade_sq], axis=0)],
        keys=["trade"],
        names=["source"],
        axis=1,
    )
    Z_net = Zx.reorder_levels(["time", "source", "country"], axis=1).sort_index(axis=1)

    Z_indices = {key: getattr(Z_net, key) for key in ["index", "columns"]}
    hydro_pumped = (
        Z_net[(Z_net < 0).any(axis=1)]
        .droplevel("source", axis=0)
        .pipe(lambda df: df.droplevel(["source", "country"], axis=1))
        .loc[:, lambda df: ~(df == 0).all(axis=0)]
        .T.groupby(level=0)
        .sum()
    )
    Z_load = (
        Z_net.sum()
        .unstack(level="time")
        .sub(Z_net.loc["trade"].sum().unstack(level="time"))
        .droplevel("source")
        .sub(
            hydro_pumped.T, fill_value=0
        )  # Here we had "negative production" of hydropower (pumping) to the load
    )
    print("saving pickles")
    save_to_pickle(Z_load, Path(root) / "data" / f"{year}" / f"Z_load_{year}.pkl")
    save_to_pickle(Z_indices, Path(root) / "data" / f"{year}" / f"indices_{year}.pkl")
    save_to_pickle(Z_net, Path(root) / "data" / f"{year}" / f"Z_net_{year}.pkl")
    print("pickles saved")
    print("treating data")
    treating_data(year, n_c, n_p, t_index, Z_net, root)
    print("..all done!")


def save_to_pickle(obj, filename):
    """
    Saves an object to a pickle file.

    Args:
        obj: The object to be saved.
        filename (Path): Path to the filename where the object will be saved.

    Returns:
        None
    """
    with open(filename, "wb") as f:
        pickle.dump(obj, f)


def load_from_pickle(filename):
    """
    Loads an object from a pickle file.

    Args:
        filename (Path): Path to the filename where the object will be saved.

    Returns:
        object: The object loaded from the pickle file.
    """
    with open(filename, "rb") as f:
        return pickle.load(f)


def add_missing_elements(df, existing_elements, all_elements, axis=0, fill_value=0):
    """
    Adds missing elements (rows or columns) to a dataframe.

    Args:
        df (pd.DataFrame): The dataframe to which elements will be added.
        existing_elements (set): The set of existing elements in the dataframe.
        all_elements (set): The set of all possible elements.
        axis (int): The axis along which to add missing elements (0 for index, 1 for columns).
        fill_value: The value to fill for the missing elements.

    Returns:
        pd.DataFrame: The dataframe with missing elements added.
    """
    missing_elements = all_elements - existing_elements
    for element in missing_elements:
        if axis == 0:
            df.loc[element] = fill_value
        else:
            df[element] = fill_value
    if axis == 0:
        df.index = pd.Index(df.index, dtype=object)
    else:
        df.columns = pd.Index(df.columns, dtype=object)

    return df.sort_index(
        axis=axis, key=lambda x: x.astype(str) if x.dtype == object else x
    )


def create_block_diagonal_matrix(data_list):
    """
    Creates a block diagonal matrix from a list of data arrays.

    Args:
        data_list (list): A list of data arrays to be combined into a block diagonal matrix.

    Returns:
        pd.DataFrame: A block diagonal matrix containing the input data.
    """
    return pd.DataFrame(block_diag(*data_list))


def process_matrix(df, operation, axis=1, **kwargs):
    """
    Applies a specified operation to a dataframe or matrix.

    Args:
        df (pd.DataFrame): The dataframe or matrix to process.
        operation (str): The operation to perform (e.g., 'normalize', 'reorder_levels').
        axis (int): The axis along which the operation should be applied, if applicable.
        kwargs: Additional arguments for specific operations.

    Returns:
        pd.DataFrame: The processed dataframe or matrix.
    """
    if operation == "normalize":
        return df.div(df.sum(axis=axis), axis=axis).fillna(0)
    elif operation == "reorder_levels":
        return df.reorder_levels(kwargs["order"], axis=axis).sort_index(axis=axis)
    else:
        raise ValueError(f"Unknown operation: {operation}")


def treating_data(year, n_c, n_p, t_index, Z_net, root):
    """
    Function called from `data_processing`. This includes heavy operations, so it's advised to run this part on a server.

    Args:
        year (int): The selected year, passed from `data_processing`.
        n_c (int): The number of all countries (including missing ones) in the dataframe, passed from `data_processing`.
        t_index ():
        root (Path): location of the data.

    Returns:
        None: Doesn't return anything; everything gets saved to the directory 'data'.
    """
    Z_indices = load_from_pickle(
        Path(root) / "data" / f"{year}" / f"indices_{year}.pkl"
    )
    Z_load = load_from_pickle(Path(root) / "data" / f"{year}" / f"Z_load_{year}.pkl")
    results, results_load = calculate_results(
        year, n_c, n_p, t_index, Z_net, Z_load, root
    )
    n_c_load = len(Z_load.columns)
    filename = Path(root) / "data" / f"{year}" / f"results_light_{year}.pkl"
    results_light = process_results_light(results, filename, n_c)

    filename = Path(root) / "data" / f"{year}" / f"results_light_load_{year}.pkl"

    L = concatenate_results(results_light, Z_net)
    output = Z_net.sum().reorder_levels(["source", "country", "time"])
    L_series = process_matrix(
        L, "reorder_levels", order=["source", "country", "time"], axis=1
    )
    output = output.reindex(L_series.columns)
    filename = Path(root) / "data" / f"{year}" / f"Z_cons_{year}.pkl"
    Z_cons = calculate_Z_cons(filename, L_series, output, Z_indices)
    save_to_pickle(
        results_load, Path(root) / "data" / f"{year}" / f"Z_load_lv_{year}.pkl"
    )


def calculate_results(year, n_c, n_p, t_index, Z_net, Z_load, root):
    """
    Calculates results by inverting matrices over time.

    Args:
        year (int): The selected year.
        n_c (int): The number of countries.
        n_p (int): The number of production mix elements.
        t_index (pd.Index): The time index.
        Z_net (pd.DataFrame): The net consumption dataframe.
        Z_load (pd.DataFrame): The load dataframe.
        root (Path): location of the data.

    Returns:
        dict: The results dictionary.
    """

    def process_time_series(data, I, t_index, filename, processor):
        if filename.exists():
            results = load_from_pickle(filename)
            print(f"Results loaded from {filename.name}")
        else:
            results = dict()
            month = None
            for t in t_index:
                if t.month != month:
                    month = t.month
                    print(f"month {month}/{t.year}")
                processor(data, t, I, results)
            save_to_pickle(results, filename)
        return results

    def process_case1(data, t, I, results):
        Z_t = data[t].reindex(data.index, axis=1, fill_value=0).values
        x_t = Z_t.sum(axis=0)
        x_t[x_t == 0] = 1  # prevent division by zero
        A_t_sparse = csr_matrix(Z_t / x_t)
        M = csr_matrix(I) - A_t_sparse
        X = spsolve(M, identity(I.shape[0], format="csc", dtype="float32"))
        results[t] = csr_matrix(X)

    def apply_load_losses(Z_load, loss_factor=0.965):
        """Applies a fixed loss factor to the entire Z_load matrix at once."""
        return Z_load * loss_factor

    # case 1: process Z_net
    filename = Path(root) / "data" / f"{year}" / f"cons_results_{year}.pkl"
    I_net = np.eye((n_p + 1) * n_c)
    print("Calculating consumption losses")
    results = process_time_series(Z_net, I_net, t_index, filename, process_case1)
    print("Calculating consumption losses finished")

    # case 2: process Z_load
    filename_load = Path(root) / "data" / f"{year}" / f"load_results_{year}.pkl"
    print("Applying load losses")
    results_load = apply_load_losses(Z_load)
    save_to_pickle(results_load, filename_load)
    print("Load losses applied and saved")

    return results, results_load


def process_results_light(results, filename, n_c):
    """
    Processes the results to a lighter format.

    Args:
        results (dict): The results dictionary.
        n_c (int): The number of countries.
        root (Path): location of the data.

    Returns:
        dict: The light results dictionary.
    """
    if filename.exists():
        results_light = load_from_pickle(filename)
        print("Light results loaded")
    else:
        print("Results light computation started")
        results_light = {}

        for k, v in results.items():
            if isinstance(v, np.ndarray):
                results_light[k] = v.astype("float32")[:, -n_c:]
            else:
                results_light[k] = csc_matrix(v[:, -n_c:], dtype="float32")

        print("Results light computation finished")
    save_to_pickle(results_light, filename)
    return results_light


def concatenate_results(results, Z):
    """
    Concatenates results into a single DataFrame.

    Args:
        results_light (dict): The light results dictionary.
        Z (pd.DataFrame): The network dataframe.

    Returns:
        pd.DataFrame: The concatenated DataFrame.
    """
    sparse_matrices = [r for r in results.values()]
    L_sparse = sp.hstack(sparse_matrices)
    num_time_steps = len(results)
    num_sub_columns = L_sparse.shape[1] // num_time_steps
    time_labels = list(results.keys())
    sub_labels = list(range(num_sub_columns)) * num_time_steps
    multi_index = pd.MultiIndex.from_tuples(
        [(time, sub) for time in time_labels for sub in range(num_sub_columns)],
        names=["time", "index"],
    )
    L_df = pd.DataFrame.sparse.from_spmatrix(L_sparse, columns=multi_index)
    L_df.columns = Z.columns
    L_df.index = Z.index
    return L_df


def calculate_Z_cons(filename, L_series, output, Z_indices):
    """
    Calculates the consumption matrix Z_cons.

    Args:
        year (int): The selected year.
        L_series (pd.DataFrame): The L series DataFrame.
        output (pd.Series): The output series.
        Z_indices (dict): The Z indices dictionary.
        root (Path): location of the data.

    Returns:
        pd.DataFrame: The consumption matrix Z_cons.
    """
    L_array = L_series.to_numpy()
    output_array = output.to_numpy().reshape(1, -1)
    Z_cons_array = L_array * output_array
    Z_cons = pd.DataFrame(Z_cons_array, index=L_series.index, columns=L_series.columns)
    if filename.exists():
        Z_cons_sp = load_from_pickle(filename)
        Z_cons = pd.DataFrame(
            Z_cons_sp.todense(), index=Z_indices["index"], columns=Z_indices["columns"]
        )
    else:
        Z_cons_sp = coo_array(
            process_matrix(
                Z_cons, "reorder_levels", order=["time", "source", "country"], axis=1
            )
        )
        save_to_pickle(Z_cons_sp, filename)
    return Z_cons
